@@ -4,8 +4,10 @@ use crate::internal::*;
 use crate::ops;
 use crate::ops::matmul::*;
 
+use super::mir_quant_unary::QMatMulUnary;
+
 #[derive(Debug, Clone, Hash, PartialEq)]
-pub struct QParams {
+pub struct MatMulQParams {
     pub a0: AttrOrInput,
     pub a_scale: AttrOrInput,
     pub b0: AttrOrInput,
@@ -14,9 +16,9 @@ pub struct QParams {
     pub c_scale: AttrOrInput,
 }
 
-impl QParams {
-    pub fn noop_static(dt: DatumType) -> QParams {
-        QParams {
+impl MatMulQParams {
+    pub fn noop_static(dt: DatumType) -> MatMulQParams {
+        MatMulQParams {
             a0: AttrOrInput::Attr(Tensor::zero_scalar_dt(dt).unwrap().into_arc_tensor()),
             a_scale: AttrOrInput::Attr(rctensor0(1f32)),
             b0: AttrOrInput::Attr(Tensor::zero_scalar_dt(dt).unwrap().into_arc_tensor()),
@@ -26,8 +28,8 @@ impl QParams {
         }
     }
 
-    pub fn all_dynamic(offset: usize) -> QParams {
-        QParams {
+    pub fn all_dynamic(offset: usize) -> MatMulQParams {
+        MatMulQParams {
             a0: AttrOrInput::Input(offset),
             a_scale: AttrOrInput::Input(offset + 1),
             b0: AttrOrInput::Input(offset + 2),
@@ -65,7 +67,7 @@ impl QParams {
         &self,
         model: &TypedModel,
         node: &TypedNode,
-    ) -> TractResult<Option<(Vec<OutletId>, QParams)>> {
+    ) -> TractResult<Option<(Vec<OutletId>, MatMulQParams)>> {
         let mut new = self.clone();
         let mut inputs = vec![];
         for (ix, input) in node.inputs.iter().enumerate() {
@@ -111,7 +113,7 @@ pub struct QMatMul {
     pub b_trans: bool,
     pub c_trans: bool,
     pub output_type: DatumType,
-    pub params: QParams,
+    pub params: MatMulQParams,
 }
 
 impl_dyn_hash!(QMatMul);
@@ -152,7 +154,7 @@ impl EvalOp for QMatMul {
         let mut model = TypedModel::default();
         let a = model.add_const("source_a", inputs[0].clone())?;
         let b = model.add_const("source_b", inputs[1].clone())?;
-        let c = model.add_const("source_c", inputs[2].clone())?;
+        let bias = model.add_const("source_bias", inputs[2].clone())?;
 
         let params = self
             .params
@@ -162,7 +164,21 @@ impl EvalOp for QMatMul {
             })
             .collect::<TractResult<Vec<_>>>()?;
 
-        let result = self.wire(&mut model, "adhoc", a, b, c, &params)?;
+        let new_op = MatMul { a_trans: self.a_trans, b_trans: self.b_trans, c_trans: self.c_trans };
+        let result = model.wire_node("adhoc.matmul", new_op, &[a, b])?[0];
+        let result = wire_matmul_quant(
+            &mut model,
+            "adhoc",
+            a,
+            self.a_trans,
+            b,
+            self.b_trans,
+            Some(bias),
+            self.c_trans,
+            result,
+            self.output_type,
+            &params,
+        )?;
         model.set_output_outlets(&[result])?;
         model.into_runnable()?.run(tvec![])
     }
@@ -191,6 +207,18 @@ impl TypedOp for QMatMul {
             self.b_trans,
             self.c_trans,
         )?;
+
+        if inputs[2].rank() == 2 {
+            let expected_bias_shape: [TDim; 2] = if self.c_trans {
+                [1.to_dim(), c_shape[c_shape.len() - 1].clone()]
+            } else {
+                [c_shape[c_shape.len() - 2].clone(), 1.to_dim()]
+            };
+            anyhow::ensure!(*inputs[2].shape == expected_bias_shape);
+        } else {
+            anyhow::ensure!(inputs[2].shape.iter().product::<TDim>() == 1.to_dim());
+        };
+
         Ok(tvec!(TypedFact::dt_shape(self.output_type, c_shape)))
     }
 
@@ -198,6 +226,94 @@ impl TypedOp for QMatMul {
         &self,
         model: &TypedModel,
         node: &TypedNode,
+    ) -> TractResult<Option<TypedModelPatch>> {
+        let a_fact = model.outlet_fact(node.inputs[0])?;
+        let b_fact = model.outlet_fact(node.inputs[1])?;
+        let bias_fact = model.outlet_fact(node.inputs[2])?;
+
+        if bias_fact.konst.is_none() {
+            return Ok(None);
+        }
+
+        let konst_ix = if a_fact.konst.is_some() {
+            0
+        } else if b_fact.konst.is_some() {
+            1
+        } else {
+            return Ok(None);
+        };
+
+        let flip = konst_ix == 1;
+        let t_konst = [self.a_trans, self.b_trans][konst_ix] ^ flip;
+        let t_var = [self.b_trans, self.a_trans][konst_ix] ^ flip;
+        let konst = model.outlet_fact(node.inputs[konst_ix])?.konst.clone().unwrap();
+        let bias = model.outlet_fact(node.inputs[2])?.konst.clone().unwrap();
+
+        let inputs: Vec<_> = node
+            .inputs
+            .iter()
+            .enumerate()
+            .filter_map(|(i, out_id)| if i == konst_ix || i == 2 { None } else { Some(*out_id) })
+            .collect();
+
+        let new_params = {
+            let mut qp = self.params.clone();
+            //compensate for the removed parameter
+            for (_, a) in qp.iter_mut() {
+                if let AttrOrInput::Input(i) = a {
+                    *i -= 2
+                }
+            }
+            if flip {
+                MatMulQParams {
+                    a0: qp.b0,
+                    a_scale: qp.b_scale,
+                    b0: qp.a0,
+                    b_scale: qp.a_scale,
+                    c0: qp.c0,
+                    c_scale: qp.c_scale,
+                }
+            } else {
+                qp
+            }
+        };
+
+        TypedModelPatch::replace_single_op(
+            model,
+            node,
+            &inputs,
+            QMatMulUnary::new(
+                konst,
+                // if bias is uniformly zero, it can be discarded
+                Some(bias).filter(|b| {
+                    b.as_uniform()
+                        .map(|b| b.cast_to_scalar::<f32>().unwrap() != 0.0)
+                        .unwrap_or(true)
+                }),
+                t_konst,
+                t_var,
+                self.c_trans ^ flip,
+                self.output_type,
+                new_params,
+            ),
+        )
+        .map(Some)
+    }
+
+    fn cost(&self, inputs: &[&TypedFact]) -> TractResult<TVec<(Cost, TDim)>> {
+        cost(
+            &inputs[0].shape.to_tvec(),
+            &inputs[1].shape.to_tvec(),
+            inputs[0].datum_type,
+            self.a_trans,
+            self.b_trans,
+        )
+    }
+
+    fn codegen(
+        &self,
+        model: &TypedModel,
+        node: &TypedNode,
     ) -> TractResult<Option<TypedModelPatch>> {
         let mut patch = TypedModelPatch::default();
 
@@ -226,82 +342,83 @@ impl TypedOp for QMatMul {
             })
             .collect::<TractResult<Vec<OutletId>>>()?;
 
-        let result = self.wire(&mut patch, &node.name, a, b, bias, &params)?;
+        let new_op = MatMul { a_trans: self.a_trans, b_trans: self.b_trans, c_trans: self.c_trans };
+        let result = patch.wire_node(format!("{}.matmul", &node.name), new_op, &[a, b])?[0];
+        let result = wire_matmul_quant(
+            &mut patch,
+            &node.name,
+            a,
+            self.a_trans,
+            b,
+            self.b_trans,
+            Some(bias),
+            self.c_trans,
+            result,
+            self.output_type,
+            &params,
+        )?;
         patch.shunt_outside(model, node.id.into(), result)?;
         Ok(Some(patch))
     }
 
-    fn cost(&self, inputs: &[&TypedFact]) -> TractResult<TVec<(Cost, TDim)>> {
-        cost(
-            &inputs[0].shape.to_tvec(),
-            &inputs[1].shape.to_tvec(),
-            inputs[0].datum_type,
-            self.a_trans,
-            self.b_trans,
-        )
-    }
-
     as_op!();
 }
 
-impl QMatMul {
-    fn wire(
-        &self,
-        model: &mut TypedModel,
-        name: &str,
-        a: OutletId,
-        b: OutletId,
-        c: OutletId,
-        params: &[OutletId],
-    ) -> TractResult<OutletId> {
-        let a_fact = model.outlet_fact(a)?.clone();
-        let rank = a_fact.rank();
-        let k = model.outlet_fact(a)?.shape[rank - 2 + !self.a_trans as usize].clone();
+pub(crate) fn wire_matmul_quant(
+    model: &mut TypedModel,
+    name: &str,
+    a: OutletId,
+    a_trans: bool,
+    b: OutletId,
+    b_trans: bool,
+    bias: Option<OutletId>,
+    c_trans: bool,
+    mut result: OutletId,
+    output_type: DatumType,
+    params: &[OutletId],
+) -> TractResult<OutletId> {
+    let a_fact = model.outlet_fact(a)?.clone();
+    let rank = a_fact.rank();
+    let m_axis = rank - 2 + c_trans as usize;
+    let n_axis = rank - 1 - c_trans as usize;
 
-        let abc_scale = combine_scales(model, name, params[1], params[3], params[5])?;
-
-        let a_i32 = model.wire_node(
-            format!("{}.a_as_i32", name),
-            ops::cast::cast(i32::datum_type()),
-            &[a],
-        )?[0];
-        let b_i32 = model.wire_node(
-            format!("{}.b_as_i32", name),
-            ops::cast::cast(i32::datum_type()),
-            &[b],
-        )?[0];
-        let a_k_axis = rank - 2 + !self.a_trans as usize;
-        let sum_a = model.wire_node(
-            format!("{}.sum_a", name),
-            ops::nn::Reduce::new(tvec!(a_k_axis), ops::nn::Reducer::Sum),
-            &[a_i32],
-        )?[0];
-        let sum_a =
-            model.wire_node(format!("{}.sum_a_reduced", name), AxisOp::Rm(a_k_axis), &[sum_a])?[0];
-        let b_k_axis = rank - 2 + self.b_trans as usize;
-        let sum_b = model.wire_node(
-            format!("{}.sum_b", name),
-            ops::nn::Reduce::new(tvec!(b_k_axis), ops::nn::Reducer::Sum),
-            &[b_i32],
-        )?[0];
-        let sum_b =
-            model.wire_node(format!("{}.sum_b_reduced", name), AxisOp::Rm(b_k_axis), &[sum_b])?[0];
-
-        let new_op = MatMul { a_trans: self.a_trans, b_trans: self.b_trans, c_trans: self.c_trans };
-        let result = model.wire_node(format!("{}.matmul", &name), new_op, &[a, b])?[0];
-        let result = wire_with_rank_broadcast(
-            &format!("{}.c", &name),
+    if let Some(bias) = bias {
+        result = wire_with_rank_broadcast(
+            &format!("{}.add_bias", &name),
             model,
             ops::math::add::bin_typed(),
-            &[result, c],
+            &[result, bias],
         )?[0];
-        let m_axis = rank - 2 + self.c_trans as usize;
-        let n_axis = rank - 1 - self.c_trans as usize;
-        let result = compensate_zero_points(
-            model, name, result, k, params[0], params[2], sum_a, sum_b, m_axis, n_axis,
-        )?;
-        requant(model, name, result, self.output_type, abc_scale, params[4])
     }
+
+    let k = model.outlet_fact(a)?.shape[rank - 2 + !a_trans as usize].clone();
+
+    let abc_scale = combine_scales(model, name, params[1], params[3], params[5])?;
+
+    let a_i32 =
+        model.wire_node(format!("{}.a_as_i32", name), ops::cast::cast(i32::datum_type()), &[a])?[0];
+    let b_i32 =
+        model.wire_node(format!("{}.b_as_i32", name), ops::cast::cast(i32::datum_type()), &[b])?[0];
+    let a_k_axis = rank - 2 + !a_trans as usize;
+    let sum_a = model.wire_node(
+        format!("{}.sum_a", name),
+        ops::nn::Reduce::new(tvec!(a_k_axis), ops::nn::Reducer::Sum),
+        &[a_i32],
+    )?[0];
+    let sum_a =
+        model.wire_node(format!("{}.sum_a_reduced", name), AxisOp::Rm(a_k_axis), &[sum_a])?[0];
+    let b_k_axis = rank - 2 + b_trans as usize;
+    let sum_b = model.wire_node(
+        format!("{}.sum_b", name),
+        ops::nn::Reduce::new(tvec!(b_k_axis), ops::nn::Reducer::Sum),
+        &[b_i32],
+    )?[0];
+    let sum_b =
+        model.wire_node(format!("{}.sum_b_reduced", name), AxisOp::Rm(b_k_axis), &[sum_b])?[0];
+    let result = compensate_zero_points(
+        model, name, result, k, params[0], params[2], sum_a, sum_b, m_axis, n_axis,
+    )?;
+    requant(model, name, result, output_type, abc_scale, params[4])
 }
 
 pub(crate) fn combine_scales(
@@ -351,6 +468,7 @@ pub(crate) fn compensate_zero_points(
 
     let sum_b =
         model.wire_node(format!("{}.reshape_sum_b", name), AxisOp::Add(m_axis), &[sum_b])?[0];
+
     debug_assert_eq!(
         model.outlet_fact(sum_a)?.shape[m_axis],
         model.outlet_fact(result)?.shape[m_axis]
@@ -463,10 +581,16 @@ pub(crate) fn clamp_and_cast_to(
         return Ok(wire);
     }
     let rank = model.outlet_fact(wire)?.rank();
-    let inf = tensor0(if dt == DatumType::I8 { -128i32 } else { 0 })
+    let inf = dt
+        .min_value()
+        .cast_to_dt(DatumType::I32)?
+        .into_owned()
         .broadcast_into_rank(rank)?
         .into_arc_tensor();
-    let sup = tensor0(if dt == DatumType::I8 { 127i32 } else { 255 })
+    let sup = dt
+        .max_value()
+        .cast_to_dt(DatumType::I32)?
+        .into_owned()
         .broadcast_into_rank(rank)?
         .into_arc_tensor();
     let wire = model.wire_node(format!("{}.min", name), ops::math::min::unary(sup), &[wire])?;
@@ -494,7 +618,7 @@ mod test {
         QMatMulProblem {
             a: arr2(&[[0]]),
             b: arr2(&[[0]]),
-            bias: arr2(&[[0]]),
+            bias: tensor0(0i32),
             a0: 0,
             b0: 0,
             c0: 1,
@@ -510,7 +634,7 @@ mod test {
         QMatMulProblem {
             a: arr2(&[[0]]),
             b: arr2(&[[0]]),
-            bias: arr2(&[[0]]),
+            bias: tensor0(0i32),
             a0: 0,
             b0: 0,
             c0: 1,
@@ -526,7 +650,7 @@ mod test {
         QMatMulProblem {
             a: arr2(&[[0]]),
             b: arr2(&[[34]]),
-            bias: arr2(&[[0]]),
+            bias: tensor0(0i32),
             a0: -17,
             b0: 1,
             c0: 0,
@@ -542,7 +666,7 @@ mod test {
         QMatMulProblem {
             a: arr2(&[[26]]),
             b: arr2(&[[0]]),
-            bias: arr2(&[[0]]),
+            bias: tensor0(0i32),
             a0: 27,
             b0: -1,
             c0: 1,
@@ -558,7 +682,7 @@ mod test {
         QMatMulProblem {
             a: arr2(&[[-23]]),
             b: arr2(&[[-2]]),
-            bias: arr2(&[[0]]),
+            bias: tensor0(0i32),
             a0: -11,
             b0: -45,
             c0: 0,
@@ -574,7 +698,7 @@ mod test {
         QMatMulProblem {
             a: arr2(&[[47], [0]]),
             b: arr2(&[[1, 0, 30]]),
-            bias: arr2(&[[0]]),
+            bias: tensor0(0i32),
             a0: 86,
             b0: 19,
             c0: 0,
@@ -590,7 +714,7 @@ mod test {
         QMatMulProblem {
             a: arr2(&[[-30]]),
             b: arr2(&[[0, 107, 0]]),
-            bias: arr2(&[[0]]),
+            bias: tensor0(0i32),
             a0: -59,
             b0: 117,
             c0: 0,
@@ -605,7 +729,7 @@ mod test {
         QMatMulProblem {
             a: arr2(&[[11, 7, 3], [10, 6, 2], [9, 5, 1], [8, 4, 0]]),
             b: arr2(&[[1, 4], [2, 5], [3, 6]]),
-            bias: arr2(&[[0]]),
+            bias: tensor0(0i32),
             a0: 12,
             b0: 0,
             c0: 0,
@@ -620,7 +744,7 @@ mod test {
     struct QMatMulProblem {
         a: Array2<i8>,
         b: Array2<i8>,
-        bias: Array2<i32>,
+        bias: Tensor,
         a0: i8,
         b0: i8,
         c0: i8,
@@ -671,13 +795,7 @@ mod test {
             );
             inputs.push(
                 model
-                    .add_source(
-                        "bias",
-                        TypedFact::dt_shape(
-                            i32::datum_type(),
-                            &[self.bias.nrows(), self.bias.ncols()],
-                        ),
-                    )
+                    .add_source("bias", TypedFact::dt_shape(i32::datum_type(), self.bias.shape()))
                     .unwrap(),
             );
             inputs.push(model.add_source("a0", TypedFact::scalar::<i8>()).unwrap());
@@ -689,7 +807,13 @@ mod test {
             let result = model
                 .wire_node(
                     "qmm",
-                    QMatMul::new(false, false, false, i8::datum_type(), QParams::all_dynamic(3)),
+                    QMatMul::new(
+                        false,
+                        false,
+                        false,
+                        i8::datum_type(),
+                        MatMulQParams::all_dynamic(3),
+                    ),
                     &inputs,
                 )
                 .unwrap();
@@ -700,7 +824,7 @@ mod test {
                 .run(tvec!(
                     self.a.clone().into_tensor(),
                     self.b.clone().into_tensor(),
-                    self.bias.clone().into_tensor(),
+                    self.bias.clone(),
                     self.a0.into(),
                     self.a_scale.into(),
                     self.b0.into(),
@@ -745,7 +869,7 @@ mod test {
                     QMatMulProblem {
                         a: Array2::from_shape_vec((m, k), a).unwrap(),
                         b: Array2::from_shape_vec((k, n), b).unwrap(),
-                        bias: arr2(&[[0i32]]),
+                        bias: tensor0(0i32),
                         a0,
                         b0,
                         c0,
